#define NEED_CPU_REG_SHORTCUTS 1
#include "bochs.h"
#define LOG_THIS BX_CPU_THIS_PTR

#include "taint/mydebug.h"
#include "taint/globals.h"

#define PINTOS

#define STACK_DEPTH 256

#if BX_USE_CPU_SMF
#define this (BX_CPU(0))
#endif

#define MASK(SHIFT, CNT) (((1ul << (CNT)) - 1) << (SHIFT))

/* Page offset (bits 0:13). */
#define PGSHIFT 0                       /* Index of first offset bit. */

#ifdef LINUX
#define PGBITS  13                      /* Number of offset bits. */
#else
#ifdef PINTOS
#define PGBITS 12
#else
#define PGBITS  13                      /* Number of offset bits. */
#warning "Dont know whether compiling for Pintos or Linux. Assuming Linux (PGBITS=13)"
#endif
#endif

#define PGMASK  MASK(PGSHIFT, PGBITS)   /* Page offset bits (0:12). */
#define PGSIZE  (1 << PGBITS)           /* Bytes in a page. */

struct stack {
    Bit32u arr[STACK_DEPTH];
    unsigned int top;
} savedRegsStack;
bool savedRegsStackInitialized = false;

void push(struct stack *s, Bit32u val) {
    assert(s->top>=0);
    if (s->top++==STACK_DEPTH) {
	DBG(ERR,("Stack Overflow. Exiting.."));
	exit(1);
    }
    s->arr[s->top-1] = val;
}

Bit32u pop(struct stack *s) {
    Bit32u ret;
    assert(s->top>=0);
    if (s->top==0) {
	DBG(ERR,("Stack Underflow. Exiting.."));
	exit(1);
    }
    ret = s->arr[s->top-1];
    s->top--;
    return ret;
}

Bit32u stack_init(struct stack *s) {
    s->top = 0;
}

void BX_CPU_C::TT_TaintSaveRegs(bxInstruction_c *i) {
	Bit32u opId = i->Id();

	if (!savedRegsStackInitialized) {
	    stack_init(&savedRegsStack);
	    savedRegsStackInitialized = true;
	}
	//if (opId==999) mylog(D1,("%s %d: called with opId 999.\n",__func__,__LINE__));
	push(&savedRegsStack, EAX);
	push(&savedRegsStack, EBX);
	push(&savedRegsStack, ECX);
	push(&savedRegsStack, EDX);
	//DBG(L1,("pushing EAX=%x, EBX=%x, ECX=%x, EDX=%x.\n",EAX,EBX,ECX,EDX));
	//g_instruction_display_count = 10;
}

void BX_CPU_C::TT_TaintRestoreRegs(bxInstruction_c *i) {
	assert(savedRegsStackInitialized);
	//mylog(D1,("%s %d: ECX=%x, EDX=%x\n",__func__,__LINE__,ECX,EDX));
	EDX = pop(&savedRegsStack);
	ECX = pop(&savedRegsStack);
	EBX = pop(&savedRegsStack);
	EAX = pop(&savedRegsStack);
	//DBG(L1,("popping EAX=%x, EBX=%x, ECX=%x, EDX=%x.\n",EAX,EBX,ECX,EDX));
}

Bit32u
BX_CPU_C::thread_current(void) {
    unsigned kernelPL = 0;
    Bit32u pid;
    if (CPL==kernelPL) {
	pid = ((ESP-1) - 0xc0000000) >> 12;
    } else {
        pid = 0;
    }
    return pid;
}

char *BX_CPU_C::backtrace(char *s) {
    Bit32u ebp, eip;
    int stackdepth = 0, readsuccessful = 1;
    char tmp[17];
    //printf("EIP = %x.\n",EIP);
    snprintf(s,16,"%x ",EIP);
    ebp = EBP;
    //while (ebp>0xc0000000 && stackdepth<10 && readsuccessful) {
    while (ebp>0xc0000000 && stackdepth<10) {
	readsuccessful &= read_virtual_dword_silent(BX_SEG_REG_SS, ebp+4, &eip);
	readsuccessful &= read_virtual_dword_silent(BX_SEG_REG_SS, ebp, &ebp);
        snprintf(tmp,16,"%x ",eip);
	s = strncat(s,tmp,16);
	stackdepth++;
    }
    return s;
}

static Bit32u
read_dword(Bit32u laddr) 
{
    Bit32u paddr = laddr - PHYS_BASE;
    if (paddr <= BX_MEM(0)->len - 4) {
        Bit32u tmp;
        ReadHostDWordFromLittleEndian(BX_MEM(0)->vector + paddr, tmp);
        return tmp;
    } else {
        return 0;
    }
}

static void
read_2_dwords(Bit32u laddr, Bit32u *dst0, Bit32u *dst1) 
{
    Bit32u paddr = laddr - PHYS_BASE;
    if (paddr <= BX_MEM(0)->len - 8) {
        ReadHostDWordFromLittleEndian(BX_MEM(0)->vector + paddr, *dst0);
        ReadHostDWordFromLittleEndian(BX_MEM(0)->vector + paddr + 4, *dst1);
    } else {
        *dst0 = 0;
        *dst1 = 0;
    }
}

void BX_CPU_C::backtrace_eips(Bit32u *eips, int n)  {
    Bit32u ebp, eip;
    *eips++ = EIP;
    ebp = EBP;
    while (ebp>0xc0000000 && --n > 0) {
        read_2_dwords(ebp, &ebp, &eip);
/* 
        eip = read_dword (ebp+4);
        ebp = read_dword (ebp);
*/
        *eips++ = eip;
    }
    while (--n > 0) {
        *eips++ = 0;
    }
}

Bit32u BX_CPU_C::callingEIP(void) {
    Bit32u ebp, eip;
    int stackdepth = 0, readsuccessful = 1;
    char tmp[17];
    ebp = EBP;
    readsuccessful = read_virtual_dword_silent(BX_SEG_REG_SS, ebp+4, &eip);
    if (readsuccessful) return eip;
    else return 0;
}

void checkbochs_log (const char *fmt, ...)
{
	va_list ap;

	if (!g_logfp) {
	    return ;
	}
	va_start (ap, fmt) ;
	vfprintf (g_logfp, fmt, ap) ;
	va_end (ap) ;
}
