#define NEED_CPU_REG_SHORTCUTS 1
#include "bochs.h"
#define LOG_THIS BX_CPU_THIS_PTR

#if BX_USE_CPU_SMF
#define this (BX_CPU(0))
#endif

#include <gdbm.h>
#include "mydebug.h"
#include "taint_type.h"
#include "lockset.h"
#include "hash.h"
#include "eraser.h"
#include "globals.h"
#include "mydebug.h"

/* Backtrace database file name. */
char btdb_fn[] = "/tmp/eraserXXXXXX";

/* Backtrace database. */
GDBM_FILE btdb;

struct backtrace_key {
    Bit32u laddr;
    unsigned seq;
};

struct backtrace {
    char type[8];
    Bit32u thread;              /* Thread ID. */
    Bit32u eips[16];            /* Backtrace EIPs. */
    Bit32u locks[16];           /* Locks held. */
};

static void
get_key(Bit32u address, unsigned seq, struct backtrace_key *key)
{
    key->laddr = address;
    key->seq = seq;
}

static datum
mk_datum(void *block, size_t length) 
{
    datum d;
    d.dptr = (char *) block;
    d.dsize = length;
    return d;
}

static void
record_backtrace(const char *type, Bit32u laddr) 
{
    lockset &tv = *BX_CPU(0)->access_linear_taint(laddr);
    struct backtrace_key key;
    struct backtrace bt;

    /* Fill in backtrace struct. */
    strncpy(bt.type, type, sizeof bt.type);
    bt.type[sizeof bt.type - 1] = '\0';
    bt.thread = BX_CPU(0)->thread_current();
    BX_CPU(0)->backtrace_eips(bt.eips, sizeof bt.eips / sizeof *bt.eips);
    lockset_dump(cur_held(bt.thread),
                 bt.locks, sizeof bt.locks / sizeof *bt.locks);

    /* Store backtrace struct. */
    get_key(laddr, tv.seq, &key);
    if (gdbm_store(btdb, mk_datum(&key, sizeof key), mk_datum(&bt, sizeof bt),
                   GDBM_REPLACE) != 0) {
        DBG(ERR, ("gdbm_store failed"));
        exit(1);
    }

    /* Increment sequence number. */
    tv.seq++;
}

static void
dump_backtrace(Bit32u laddr, unsigned seq)
{
    struct backtrace_key key;
    struct backtrace *bt;
    datum content;

    /* Fetch backtrace. */
    get_key(laddr, seq, &key);
    content = gdbm_fetch(btdb, mk_datum(&key, sizeof key));
    if (content.dptr == NULL) {
        abort();
    }
    bt = (backtrace *) content.dptr;

    /* Dump. */
    fprintf(g_logfp, "%s: thread %x, backtrace", bt->type, bt->thread);
    for (unsigned i = 0; i < sizeof bt->eips / sizeof *bt->eips; i++) {
        if (bt->eips[i] == 0)
            break;
        fprintf(g_logfp, " %x", bt->eips[i]); 
    }
    fprintf(g_logfp, ", locks");
    for (unsigned i = 0; i < sizeof bt->locks / sizeof *bt->locks; i++) {
        if (bt->locks[i] == 0)
            break;
        fprintf(g_logfp, " %x", bt->locks[i]);
    }
    fprintf(g_logfp, "\n");

    free(bt);
}

    
    


void breakme() {}

#define WRN_UNINIT(loc) do {  \
    /* DBG(WRN,("Thread %x: Read on uninitialized location %x, backtrace: %s\n",myid,loc, backtrace(btstr))); */ \
} while (0)

#define WRN_ERASER(myid,loc,tval) do {  \
    if (!already_warned(loc)) { \
        warn(loc); \
	DBG(WRN,("Thread %x: Warning on location %x, backtrace: %s\n",myid,loc, backtrace(btstr))); \
	breakme(); \
    } \
} while (0)

struct warn_table_entry {
    unsigned loc;
    hash_elem h_elem;
};

static int warn_table_initialized = 0;
static struct hash warn_table ;

unsigned warn_hash (const hash_elem *e, void *aux)
{
    struct warn_table_entry *h = hash_entry (e, struct warn_table_entry, h_elem);
    return h->loc;
}

bool
warn_less (const hash_elem *a_, const hash_elem *b_,
			void *aux)
{
    struct warn_table_entry *a = hash_entry (a_, struct warn_table_entry, h_elem);
    struct warn_table_entry *b = hash_entry (b_, struct warn_table_entry, h_elem);
    return (a->loc < b->loc);
}

static int
already_warned(unsigned loc) {
    struct warn_table_entry tmp;
    hash_elem *h_element;
    if (!warn_table_initialized) return 0;
    tmp.loc = loc;
    h_element = hash_find(&warn_table, &tmp.h_elem);
    if (h_element) return 1;
    else return 0;
}

static int
warn(unsigned loc) {
    struct warn_table_entry *tmp = (struct warn_table_entry*)malloc(sizeof(struct warn_table_entry));
    if (!warn_table_initialized) {
	hash_init(&warn_table, warn_hash, warn_less, NULL);
	warn_table_initialized = 1;
    }
    tmp->loc = loc;
    hash_insert(&warn_table, &tmp->h_elem);
}


void BX_CPU_C::TT_Lock(bxInstruction_c *i) {
    Bit32u opId = i->Id();
    if (opId!=ERASER_ID) return;

    Bit32u myid = (BX_CPU_THIS_PTR thread_current())&0x3fffffff;
    DBG(LOCKS,("%x: acquiring lock %x. backtrace: %s\n",myid,ECX,backtrace(btstr)));
    eraser_lock(ECX);
    //lockset_t lset = add_lock(cur_held(myid),ECX);
    //update_lockset(myid,lset);
}

void BX_CPU_C::TT_Unlock(bxInstruction_c *i) {
    Bit32u opId = i->Id();
    if (opId!=ERASER_ID) return;

    Bit32u myid = (BX_CPU_THIS_PTR thread_current())&0x3fffffff;
    DBG(LOCKS,("%x: releasing lock %x. backtrace: %s\n",myid,ECX,backtrace(btstr)));
    eraser_unlock(ECX);
    //Bit32u myid = (BX_CPU_THIS_PTR thread_current())&0x3fffffff;
    //lockset_t lset = remove_lock(cur_held(myid),ECX);
    //update_lockset(myid,lset);
}

void BX_CPU_C::eraser_access_linear(bx_address laddr, unsigned len, unsigned pl, unsigned rw, void *notused)
{
    Bit32u tid = BX_CPU_THIS_PTR thread_current ();
    if (ignore_on (tid))
        return;

    // Acquire a dummy lock for disabled interrupts.
    if (!BX_CPU_THIS_PTR get_IF())
    	eraser_lock(INTERRUPT_LOCK);

    // Acquire a dummy lock for h/w prefix "LOCK".
    if (BX_CPU_THIS_PTR curInstruction->isLocked())
    	eraser_lock(HW_PREFIX_LOCK);

    for (; len-- > 0; laddr++) {
        lockset *tvp = access_linear_taint(laddr);
        if (tvp == NULL)
            continue;
        lockset &tv = *tvp;
        lockset orig_tv = tv;

        switch (tv.state) {
        case S_VIRGIN:
	    if (rw==BX_WRITE || rw==BX_RW) {
                tv.state = S_EXCLUSIVE;
                tv.seq = 0;
                tv.value = tid;
                record_backtrace("V->E", laddr);
	    } else {
		//WRN_UNINIT(laddr) ;
	    }
            break;

        case S_EXCLUSIVE:
	    if (tv.value == tid) {
                // Still exclusive to same thread.
            } else {
		tv.value = cur_held (tid);
		if (rw==BX_WRITE || rw==BX_RW) {
		    tv.state = S_SH_MOD;
                    record_backtrace("E->SM", laddr);
		} else {
                    tv.state = S_SHARED;
                    record_backtrace("E->S", laddr);
		}
	    }
            break;

        case S_SHARED:
            tv.value = intersect_locksets (tv.value, cur_held (tid));
	    if (rw==BX_WRITE || rw==BX_RW) {
		tv.state = S_SH_MOD;
                record_backtrace("S->SM", laddr);
	    } else if (tv.value != orig_tv.value) {
                record_backtrace("S", laddr);
	    }
            break;

        case S_SH_MOD:
            tv.value = intersect_locksets (tv.value, cur_held (tid));
            if (tv.value != orig_tv.value) {
                record_backtrace("SM", laddr);
	    }
	}

	/* Warn if needed */
	if (tv.state == S_SH_MOD
            && tv.value == LOCKSET_EMPTY
            && !already_warned(laddr)) {
            warn(laddr);
            
            unsigned v_e_seq;
            
            fprintf(g_logfp, "Warning on location %x:\n", laddr);

            for (unsigned i = 0; i < tv.seq; i++)
                dump_backtrace(laddr, i);

            fprintf(g_logfp, "\n");
        }
    }

    // Release interrupt lock. 
    if (!BX_CPU_THIS_PTR get_IF())
        eraser_unlock(INTERRUPT_LOCK);

    // Release LOCK prefix lock.
    if (BX_CPU_THIS_PTR curInstruction->isLocked())
	eraser_unlock(HW_PREFIX_LOCK);
}

void BX_CPU_C::TT_CommonOps(bxInstruction_c *i) {
    Bit32u opId = i->Id();
    Bit32u myid = (BX_CPU_THIS_PTR thread_current())&0x3fffffff;
    if (opId!=ERASER_ID) return;

    if (EAX==IGNORE_OP) {
	for (int i=0;i<ECX;i++) warn(EDX+i);
	DBG(L1,("ignoring location %x (%d).\n",EDX,ECX));
    }
    else if (EAX==LOCKINIT_OP) {
	eraser_init(ECX);
    }
    else if (EAX==IGNOREON_OP) {
	if (global_startup_ignore) {
	    return ;
	}
	set_ignore(myid,true);
	DBG(L1,("setting ignore on for thread %x. backtrace: %s\n",myid,backtrace(btstr)));
    }
    else if (EAX==IGNOREOFF_OP) {
	if (global_startup_ignore) {
	    return ;
	}
	set_ignore(myid,false);
	DBG(L1,("setting ignore off for thread %x. backtrace: %s\n",myid,backtrace(btstr)));
    }
    else if (EAX==REUSE_OP) {
	int pl = 0;	//kernel privileges
	DBG(L1,("%x: reusing location %x (%d). ESP=%x\n",myid,EDX,ECX,ESP));
	for (int i=0;i<ECX;i++) {
            lockset *tv = access_linear_taint(EDX+i);
            if (tv != NULL)
                tv->state = S_VIRGIN;
	}
    } else if (EAX==DBG_MARK_OP) {
        Bit32u taintval = 0;
	char str[MAX_STRLEN];
	int pl = 0;	//kernel privileges
	int ret, i=0;
	do {
	    ret = access_linear_silent(EDX+i,1,pl,BX_READ,&str[i]) ;
	} while (ret && str[i] && ++i<MAX_STRLEN);
	str[i] = '\0';
	DBG(L1,("%x: dbg mark at %s:%d. EIP=%x\n",myid,str,ECX,EIP));
    } else if (EAX==GLOBAL_STARTUP_IGNOREOFF_OP) {
	assert (global_startup_ignore) ;
	global_startup_ignore = false ;
	DBG(L1,("%x: setting global_startup_ignore to off\n",myid));
    }
}

void BX_CPU_C::eraser_init_globals(void) {
    g_logfp = fopen (g_logfn, "w") ;
    if (g_logfp==NULL) {
	DBG (ERR, ("%s(): Error opening checkbochs log %s for writing.\n",__func__,g_logfn)) ;
    }

    /* Create temporary file and set name into btdb_fn. */
    int old_umask = umask(0077);
    int fd = mkstemp(btdb_fn);
    if (fd < 0) {
        DBG(ERR, ("mkstemp: Couldn't create temp file: %s.\n", strerror(errno)));
        exit(1);
    }
    umask(old_umask);
    close(fd);
    
    /* Create database. */
    btdb = gdbm_open(btdb_fn, 0, GDBM_NEWDB | GDBM_NOLOCK, 0600, NULL);
    if (btdb == NULL) {
        DBG(ERR, ("gdbm_open: Couldn't create backtrace database: %s.\n",
                  gdbm_strerror(gdbm_errno)));
        exit(1);
    }
    unlink(btdb_fn);
}

void BX_CPU_C::eraser_done_globals(void) 
{
    if (global_startup_ignore) {
        fprintf(g_logfp,
                "WARNING: Eraser never enabled.  Did you really apply "
                "pintos/src/misc/checkbochs.patch?\n");
    }
}

